#!/usr/bin/env python

import os
import sys
import datetime
import argparse
import celery
import analysis_engine.consts as ae_consts
import analysis_engine.run_custom_algo as run_custom_algo
import analysis_engine.work_tasks.get_celery_app as get_celery_app
import analysis_engine.plot_trading_history as plot_trading_history
import analysis_engine.charts as ae_charts
import analysis_engine.iex.extract_df_from_redis as extract_utils
import analysis_engine.show_dataset as show_dataset
import analysis_engine.load_history_dataset_from_file as load_history
import analysis_engine.load_report_dataset_from_file as load_report
import analysis_engine.restore_dataset as restore_dataset
import analysis_engine.work_tasks.prepare_pricing_dataset as prep_dataset
import analysis_engine.api_requests as api_requests
import spylunking.log.setup_logging as log_utils


# Disable celery log hijacking
# https://github.com/celery/celery/issues/2509
@celery.signals.setup_logging.connect
def setup_celery_logging(**kwargs):
    pass


log = log_utils.build_colorized_logger(
    name='aws-backup',
    log_config_path=ae_consts.LOG_CONFIG_PATH)


def restore_missing_dataset_values_from_algo_ready_file(
        ticker,
        path_to_file,
        redis_address,
        redis_password,
        redis_db=ae_consts.REDIS_DB,
        output_redis_db=None,
        compress=False,
        encoding='utf-8',
        dataset_type=ae_consts.SA_DATASET_TYPE_ALGO_READY,
        serialize_datasets=ae_consts.DEFAULT_SERIALIZED_DATASETS,
        show_summary=True):
    """restore_missing_dataset_values_from_algo_ready_file

    restore missing dataset nodes in redis from an algorithm-ready
    dataset file on disk - use this to restore redis from scratch

    :param ticker: string ticker
    :param path_to_file: string path to file on disk
    :param redis_address: redis server endpoint adddress with
        format ``host:port``
    :param redis_password: optional - string password for redis
    :param redis_db: redis db (default is ``REDIS_DB``)
    :param output_redis_db: optional - integer for different
        redis database (default is ``None``)
    :param compress: contents in algorithm-ready file are
        compressed (default is ``False``)
    :param encoding: byte encoding of algorithm-ready file
        (default is ``utf-8``)
    :param dataset_type: optional - dataset type
        (default is ``SA_DATASET_TYPE_ALGO_READY``)
    :param serialize_datasets: optional - list of dataset names to
        deserialize in the dataset
    :param show_summary: optional - show a summary of the algorithm-ready
        dataset using ``analysis_engine.show_dataset.show_dataset``
        (default is ``True``)
    """
    if not os.path.exists(path_to_file):
        log.error(
            'missing file={} for restore'.format(
                path_to_file))
        return

    if dataset_type == ae_consts.SA_DATASET_TYPE_ALGO_READY:
        log.info(
            'restore start - load dataset from file={}'.format(
                path_to_file))
    else:
        log.error(
            'restore dataset unsupported type={} for file={}'.format(
                dataset_type,
                path_to_file))
        return

    if not output_redis_db:
        output_redis_db = redis_db

    restore_dataset.restore_dataset(
        show_summary=show_summary,
        path_to_file=path_to_file,
        compress=compress,
        encoding=encoding,
        dataset_type=dataset_type,
        serialize_datasets=serialize_datasets,
        redis_address=redis_address,
        redis_password=redis_password,
        redis_db=redis_db,
        redis_output_db=output_redis_db,
        verbose=False)
    log.info(
        'restore done - dataset in file={}'.format(
            path_to_file))
# end of restore_missing_dataset_values_from_algo_ready_file


def examine_dataset_in_file(
        path_to_file,
        compress=False,
        encoding='utf-8',
        ticker=None,
        dataset_type=ae_consts.SA_DATASET_TYPE_ALGO_READY,
        serialize_datasets=ae_consts.DEFAULT_SERIALIZED_DATASETS,):
    """examine_dataset_in_file

    Show the internal dataset dictionary structure in dataset file

    :param path_to_file: path to file
    :param compress: optional - boolean flag for decompressing
        the contents of the ``path_to_file`` if necessary
        (default is ``False`` and algorithms
        use ``zlib`` for compression)
    :param encoding: optional - string for data encoding
    :param ticker: optional - string ticker symbol
        to verify is in the dataset
    :param dataset_type: optional - dataset type
        (default is ``SA_DATASET_TYPE_ALGO_READY``)
    :param serialize_datasets: optional - list of dataset names to
        deserialize in the dataset
    """
    if dataset_type == ae_consts.SA_DATASET_TYPE_ALGO_READY:
        log.info(
            'show start - load dataset from file={}'.format(
                path_to_file))
        show_dataset.show_dataset(
            path_to_file=path_to_file,
            compress=compress,
            encoding=encoding,
            dataset_type=dataset_type,
            serialize_datasets=serialize_datasets)
        log.info(
            'show done - dataset in file={}'.format(
                path_to_file))
    elif dataset_type == ae_consts.SA_DATASET_TYPE_TRADING_HISTORY:
        log.info(
            'load trading history dataset '
            'from file={}'.format(
                path_to_file))
        trading_history_dict = load_history.load_history_dataset_from_file(
            path_to_file=path_to_file,
            compress=compress,
            encoding=encoding)
        history_df = trading_history_dict[ticker]

        first_date = history_df['date'].iloc[0]
        end_date = history_df['date'].iloc[-1]
        title = (
            'Trading History {} for Algo {}\n'
            'Backtest dates from {} to {}'.format(
                ticker,
                trading_history_dict['algo_name'],
                first_date,
                end_date))
        xcol = 'date'
        xlabel = 'Dates vs {} values'.format(
            trading_history_dict['algo_name'])
        ylabel = 'Algo Values from columns:\n{}'.format(
            trading_history_dict['algo_name'],
            list(history_df.columns.values))
        df_filter = (history_df['close'] > 0.01)

        # set default hloc columns:
        red = 'close'
        blue = 'low'
        green = 'high'
        orange = 'open'

        log.info(
            'available columns to plot in dataset: {}'.format(
                ae_consts.ppj(list(history_df.columns.values))))

        plot_trading_history.plot_trading_history(
            title=title,
            df=history_df,
            red=red,
            blue=blue,
            green=green,
            orange=orange,
            date_col=xcol,
            xlabel=xlabel,
            ylabel=ylabel,
            df_filter=df_filter,
            show_plot=True,
            dropna_for_all=False)
    elif dataset_type == ae_consts.SA_DATASET_TYPE_TRADING_REPORT:
        log.info(
            'load trading performance report dataset '
            'from file={}'.format(
                path_to_file))
        trading_report_dict = load_report.load_report_dataset_from_file(
            path_to_file=path_to_file,
            compress=compress,
            encoding=encoding)
        print(trading_report_dict)
    else:
        log.error(
            'show unsupported dataset type={} for file={}'.format(
                dataset_type,
                path_to_file))
        return
# end of examine_dataset_in_file


def run_aws_backup():
    """run_aws_backup

    Run buy and sell analysis on a stock to send alerts to subscribed
    users

    """

    log.debug('start - aws-backup')

    parser = argparse.ArgumentParser(
        description=(
            'stock analysis tool'))
    parser.add_argument(
        '-t',
        help=(
            'ticker'),
        required=True,
        dest='ticker')
    parser.add_argument(
        '-e',
        help=(
            'file path to extract an '
            'algorithm-ready datasets from redis'),
        required=False,
        dest='algo_extract_loc')
    parser.add_argument(
        '-l',
        help=(
            'show dataset in this file'),
        required=False,
        dest='show_from_file')
    parser.add_argument(
        '-H',
        help=(
            'show trading history dataset in this file'),
        required=False,
        dest='show_history_from_file')
    parser.add_argument(
        '-E',
        help=(
            'show trading performance report dataset in this file'),
        required=False,
        dest='show_report_from_file')
    parser.add_argument(
        '-L',
        help=(
            'restore an algorithm-ready dataset file back into redis'),
        required=False,
        dest='restore_algo_file')
    parser.add_argument(
        '-f',
        help=(
            'run in mode: prepare dataset from '
            'redis key or s3 key'),
        required=False,
        dest='prepare_mode',
        action='store_true')
    parser.add_argument(
        '-J',
        help=(
            'plot action - after preparing you can use: '
            '-J show to open the image (good for debugging)'),
        required=False,
        dest='plot_action')
    parser.add_argument(
        '-b',
        help=(
            'run a backtest using the dataset in '
            'a file path/s3 key/redis key formats: '
            'file:/opt/sa/tests/datasets/algo/SPY-latest.json or '
            's3://algoready/SPY-latest.json or '
            'redis:SPY-latest'),
        required=False,
        dest='backtest_loc')
    parser.add_argument(
        '-B',
        help=(
            'optional - broker url for Celery'),
        required=False,
        dest='broker_url')
    parser.add_argument(
        '-C',
        help=(
            'optional - broker url for Celery'),
        required=False,
        dest='backend_url')
    parser.add_argument(
        '-w',
        help=(
            'optional - flag for publishing an algorithm job '
            'using Celery to the ae workers'),
        required=False,
        dest='run_on_engine',
        action='store_true')
    parser.add_argument(
        '-k',
        help=(
            'optional - s3 access key'),
        required=False,
        dest='s3_access_key')
    parser.add_argument(
        '-K',
        help=(
            'optional - s3 secret key'),
        required=False,
        dest='s3_secret_key')
    parser.add_argument(
        '-a',
        help=(
            'optional - s3 address format: <host:port>'),
        required=False,
        dest='s3_address')
    parser.add_argument(
        '-Z',
        help=(
            'optional - s3 secure: default False'),
        required=False,
        dest='s3_secure')
    parser.add_argument(
        '-s',
        help=(
            'optional - start date: YYYY-MM-DD'),
        required=False,
        dest='start_date')
    parser.add_argument(
        '-n',
        help=(
            'optional - end date: YYYY-MM-DD'),
        required=False,
        dest='end_date')
    parser.add_argument(
        '-u',
        help=(
            'optional - s3 bucket name'),
        required=False,
        dest='s3_bucket_name')
    parser.add_argument(
        '-G',
        help=(
            'optional - s3 region name'),
        required=False,
        dest='s3_region_name')
    parser.add_argument(
        '-g',
        help=(
            'Path to a custom algorithm module file '
            'on disk. This module must have a single '
            'class that inherits from: '
            'https://github.com/AlgoTraders/stock-analysis-engine/'
            'blob/master/'
            'analysis_engine/algo.py Additionally you '
            'can find the Example-Minute-Algorithm here: '
            'https://github.com/AlgoTraders/stock-analysis-engine/'
            'blob/master/analysis_engine/mocks/'
            'example_algo_minute.py'),
        required=False,
        dest='run_algo_in_file')
    parser.add_argument(
        '-p',
        help=(
            'optional - s3 bucket/file for trading history'),
        required=False,
        dest='algo_history_loc')
    parser.add_argument(
        '-o',
        help=(
            'optional - s3 bucket/file for trading performance report'),
        required=False,
        dest='algo_report_loc')
    parser.add_argument(
        '-r',
        help=(
            'optional - redis_address format: <host:port>'),
        required=False,
        dest='redis_address')
    parser.add_argument(
        '-R',
        help=(
            'optional - redis and s3 key name'),
        required=False,
        dest='keyname')
    parser.add_argument(
        '-m',
        help=(
            'optional - redis database number (0 by default)'),
        required=False,
        dest='redis_db')
    parser.add_argument(
        '-x',
        help=(
            'optional - redis expiration in seconds'),
        required=False,
        dest='redis_expire')
    parser.add_argument(
        '-z',
        help=(
            'optional - strike price'),
        required=False,
        dest='strike')
    parser.add_argument(
        '-c',
        help=(
            'optional - algorithm config_file path for setting '
            'up internal algorithm trading strategies and '
            'indicators'),
        required=False,
        dest='config_file')
    parser.add_argument(
        '-P',
        help=(
            'optional - get pricing data if "1" or "0" disabled'),
        required=False,
        dest='get_pricing')
    parser.add_argument(
        '-N',
        help=(
            'optional - get news data if "1" or "0" disabled'),
        required=False,
        dest='get_news')
    parser.add_argument(
        '-O',
        help=(
            'optional - get options data if "1" or "0" disabled'),
        required=False,
        dest='get_options')
    parser.add_argument(
        '-i',
        help=(
            'optional - ignore column names (comma separated)'),
        required=False,
        dest='ignore_columns')
    parser.add_argument(
        '-d',
        help=(
            'debug'),
        required=False,
        dest='debug',
        action='store_true')
    args = parser.parse_args()

    mode = 'prepare'
    plot_action = ae_consts.PLOT_ACTION_SHOW
    ticker = ae_consts.TICKER
    ticker_id = ae_consts.TICKER_ID
    ssl_options = ae_consts.SSL_OPTIONS
    transport_options = ae_consts.TRANSPORT_OPTIONS
    broker_url = ae_consts.WORKER_BROKER_URL
    backend_url = ae_consts.WORKER_BACKEND_URL
    path_to_config_module = ae_consts.WORKER_CELERY_CONFIG_MODULE
    include_tasks = ae_consts.INCLUDE_TASKS
    s3_access_key = ae_consts.S3_ACCESS_KEY
    s3_secret_key = ae_consts.S3_SECRET_KEY
    s3_region_name = ae_consts.S3_REGION_NAME
    s3_address = ae_consts.S3_ADDRESS
    s3_secure = ae_consts.S3_SECURE
    s3_bucket_name = ae_consts.S3_BUCKET
    s3_key = ae_consts.S3_KEY
    redis_address = ae_consts.REDIS_ADDRESS
    redis_key = ae_consts.REDIS_KEY
    redis_password = ae_consts.REDIS_PASSWORD
    redis_db = ae_consts.REDIS_DB
    redis_expire = ae_consts.REDIS_EXPIRE
    dataset_type = ae_consts.SA_DATASET_TYPE_ALGO_READY
    serialize_datasets = ae_consts.DEFAULT_SERIALIZED_DATASETS
    output_redis_key = None
    output_s3_bucket = None
    output_s3_key = None
    ignore_columns = None
    compress = False
    encoding = 'utf-8'
    slack_enabled = False
    slack_code_block = False
    slack_full_width = False
    verbose = False
    debug = False

    redis_serializer = 'json'
    redis_encoding = 'utf-8'
    output_redis_key = None
    output_s3_bucket = None
    output_s3_key = None
    s3_enabled = True
    redis_enabled = True
    ignore_columns = None
    debug = False

    run_on_engine = False
    show_from_file = None
    show_history_from_file = None
    show_report_from_file = None
    restore_algo_file = None
    backtest_loc = None
    use_custom_algo = False
    algo_history_loc = 's3://algohistory'
    algo_report_loc = 's3://algoreport'
    algo_extract_loc = 's3://algoready'

    use_balance = 5000.0
    use_commission = 6.0
    auto_fill = True
    use_start_date = None
    use_end_date = None
    use_config_file = None
    use_name = 'myalgo'

    if args.ticker:
        ticker = args.ticker.upper()
    if args.broker_url:
        broker_url = args.broker_url
    if args.backend_url:
        backend_url = args.backend_url
    if args.s3_access_key:
        s3_access_key = args.s3_access_key
    if args.s3_secret_key:
        s3_secret_key = args.s3_secret_key
    if args.s3_region_name:
        s3_region_name = args.s3_region_name
    if args.s3_address:
        s3_address = args.s3_address
        s3_enabled = True
    if args.s3_secure:
        s3_secure = args.s3_secure
    if args.s3_bucket_name:
        s3_bucket_name = args.s3_bucket_name
    if args.keyname:
        s3_key = args.keyname
        redis_key = args.keyname
    if args.redis_address:
        redis_address = args.redis_address
    if args.redis_db:
        redis_db = args.redis_db
    if args.redis_expire:
        redis_expire = args.redis_expire
    if args.prepare_mode:
        mode = ae_consts.SA_MODE_PREPARE
    if args.ignore_columns:
        ignore_columns_org = args.ignore_columns
        ignore_columns = ignore_columns_org.split(",")
    if args.plot_action:
        if str(args.plot_action).lower() == 'show':
            plot_action = ae_consts.PLOT_ACTION_SHOW
        elif str(args.plot_action).lower() == 's3':
            plot_action = ae_consts.PLOT_ACTION_SAVE_TO_S3
        elif str(args.plot_action).lower() == 'save':
            plot_action = ae_consts.PLOT_ACTION_SAVE_AS_FILE
        else:
            plot_action = ae_consts.PLOT_ACTION_SHOW
            log.warning(
                'unsupported plot_action: {}'.format(
                    args.plot_action))

    if args.debug:
        debug = True

    if args.algo_extract_loc:
        mode = ae_consts.SA_MODE_EXTRACT
    if args.show_from_file:
        show_from_file = args.show_from_file
        mode = ae_consts.SA_MODE_SHOW_DATASET
    if args.show_history_from_file:
        show_history_from_file = args.show_history_from_file
        mode = ae_consts.SA_MODE_SHOW_HISTORY_DATASET
    if args.show_report_from_file:
        show_report_from_file = args.show_report_from_file
        mode = ae_consts.SA_MODE_SHOW_REPORT_DATASET
    if args.restore_algo_file:
        restore_algo_file = args.restore_algo_file
        mode = ae_consts.SA_MODE_RESTORE_REDIS_DATASET
    if args.run_algo_in_file:
        mode = ae_consts.SA_MODE_RUN_ALGO
    if args.backtest_loc:
        mode = ae_consts.SA_MODE_RUN_ALGO
    if args.start_date:
        try:
            use_start_date = '{} 00:00:00'.format(
                str(args.start_date))
            datetime.datetime.strptime(
                args.start_date,
                ae_consts.COMMON_DATE_FORMAT)
        except Exception as e:
            msg = (
                'please use a start date formatted as: {}'
                '\n'
                'error was: {}'.format(
                    ae_consts.COMMON_DATE_FORMAT,
                    e))
            log.error(msg)
            sys.exit(1)
        # end of testing for a valid date
    # end of args.start_date
    if args.end_date:
        try:
            use_end_date = '{} 00:00:00'.format(
                str(args.end_date))
            datetime.datetime.strptime(
                args.end_date,
                ae_consts.COMMON_DATE_FORMAT)
        except Exception as e:
            msg = (
                'please use an end date formatted as: {}'
                '\n'
                'error was: {}'.format(
                    ae_consts.COMMON_DATE_FORMAT,
                    e))
            log.error(msg)
            sys.exit(1)
        # end of testing for a valid date
    # end of args.end_date
    if args.config_file:
        use_config_file = args.config_file
        if not os.path.exists(use_config_file):
            log.error(
                'Failed: unable to find config file: -c {}'.format(
                    use_config_file))
            sys.exit(1)

    config_dict = None
    load_from_s3_bucket = None
    load_from_s3_key = None
    load_from_redis_key = None
    load_from_file = None
    load_compress = False
    load_publish = True
    load_config = None
    report_redis_key = None
    report_s3_bucket = None
    report_s3_key = None
    report_file = None
    report_compress = False
    report_publish = False
    report_config = None
    history_redis_key = None
    history_s3_bucket = None
    history_s3_key = None
    history_file = None
    history_compress = False
    history_publish = False
    history_config = None
    extract_redis_key = None
    extract_s3_bucket = None
    extract_s3_key = None
    extract_file = None
    extract_save_dir = None
    extract_compress = False
    extract_publish = True
    extract_config = None
    publish_to_slack = False
    publish_to_s3 = True
    publish_to_redis = False
    use_timeseries = 'day'
    use_trade_strategy = 'count'

    valid = False
    required_task = False
    work = None
    task_name = None
    work = {}
    path_to_tasks = 'analysis_engine.work_tasks'
    if mode == ae_consts.SA_MODE_PREPARE:
        task_name = (
            '{}.'
            'prepare_pricing_dataset.prepare_pricing_dataset').format(
                path_to_tasks)
        work = api_requests.build_prepare_dataset_request()
        if output_s3_key:
            work['prepared_s3_key'] = output_s3_key
        if output_s3_bucket:
            work['prepared_s3_bucket'] = output_s3_bucket
        if output_redis_key:
            work['prepared_redis_key'] = output_redis_key
        work['ignore_columns'] = ignore_columns
        valid = True
        required_task = True
    elif mode == ae_consts.SA_MODE_EXTRACT:
        if args.algo_extract_loc:
            algo_extract_loc = args.algo_extract_loc
            if ('file:/' not in algo_extract_loc and
                    's3://' not in algo_extract_loc and
                    'redis://' not in algo_extract_loc):
                log.error(
                    'invalid -e <extract_to_file_or_s3_key_or_redis_key> '
                    'specified. please use either: '
                    '-e file:/opt/sa/tests/datasets/algo/SPY-latest.json or '
                    '-e s3://algoready/SPY-latest.json or '
                    '-e redis://SPY-latest')
                sys.exit(1)
            if 's3://' in algo_extract_loc:
                extract_s3_bucket = algo_extract_loc.split('/')[-2]
                extract_s3_key = algo_extract_loc.split('/')[-1]
            elif 'redis://' in algo_extract_loc:
                extract_redis_key = algo_extract_loc.split('/')[-1]
            elif 'file:/' in algo_extract_loc:
                extract_file = algo_extract_loc.split(':')[-1]
        # end of parsing supported transport for loading

        use_custom_algo = True
    elif mode == ae_consts.SA_MODE_SHOW_DATASET:
        examine_dataset_in_file(
            ticker=ticker,
            path_to_file=show_from_file)
        log.info(
            'done showing {} dataset from file={}'.format(
                ticker,
                show_from_file))
        sys.exit(0)
    elif mode == ae_consts.SA_MODE_SHOW_HISTORY_DATASET:
        examine_dataset_in_file(
            ticker=ticker,
            dataset_type=ae_consts.SA_DATASET_TYPE_TRADING_HISTORY,
            path_to_file=show_history_from_file)
        log.info(
            'done showing trading history {} dataset from '
            'file={}'.format(
                ticker,
                show_from_file))
        sys.exit(0)
    elif mode == ae_consts.SA_MODE_SHOW_REPORT_DATASET:
        examine_dataset_in_file(
            ticker=ticker,
            dataset_type=ae_consts.SA_DATASET_TYPE_TRADING_REPORT,
            path_to_file=show_report_from_file)
        log.info(
            'done showing trading performance report {} dataset from '
            'file={}'.format(
                ticker,
                show_from_file))
        sys.exit(0)
    elif mode == ae_consts.SA_MODE_RESTORE_REDIS_DATASET:
        restore_missing_dataset_values_from_algo_ready_file(
            ticker=ticker,
            path_to_file=restore_algo_file,
            redis_address=redis_address,
            redis_password=redis_password,
            redis_db=redis_db,
            output_redis_db=redis_db,
            dataset_type=ae_consts.SA_DATASET_TYPE_ALGO_READY,
            serialize_datasets=ae_consts.DEFAULT_SERIALIZED_DATASETS)
        log.info(
            'done restoring {} dataset from file={} into redis_db={}'.format(
                ticker,
                restore_algo_file,
                redis_db))
        sys.exit(0)
    elif mode == ae_consts.SA_MODE_RUN_ALGO:
        if args.run_algo_in_file:
            if not os.path.exists(args.run_algo_in_file):
                log.error(
                    'missing algorithm module file: {}'.format(
                        args.run_algo_in_file))
                sys.exit(1)

        if args.backtest_loc:
            backtest_loc = args.backtest_loc
            if ('file:/' not in backtest_loc and
                    's3://' not in backtest_loc and
                    'redis://' not in backtest_loc):
                log.error(
                    'invalid -b <backtest dataset file> specified. '
                    '{} '
                    'please use either: '
                    '-b file:/opt/sa/tests/datasets/algo/SPY-latest.json or '
                    '-b s3://algoready/SPY-latest.json or '
                    '-b redis://SPY-latest'.format(
                        backtest_loc))
                sys.exit(1)
            if 's3://' in backtest_loc:
                load_from_s3_bucket = backtest_loc.split('/')[-2]
                load_from_s3_key = backtest_loc.split('/')[-1]
            elif 'redis://' in backtest_loc:
                load_from_redis_key = backtest_loc.split('/')[-1]
            elif 'file:/' in backtest_loc:
                load_from_file = backtest_loc.split(':')[-1]
            load_publish = True
        # end of parsing supported transport - loading an algo-ready

        if args.algo_history_loc:
            algo_history_loc = args.algo_history_loc
            if ('file:/' not in algo_history_loc and
                    's3://' not in algo_history_loc and
                    'redis://' not in algo_history_loc):
                log.error(
                    'invalid -p <backtest dataset file> specified. '
                    '{} '
                    'please use either: '
                    '-p file:/opt/sa/tests/datasets/algo/SPY-latest.json or '
                    '-p s3://algoready/SPY-latest.json or '
                    '-p redis://SPY-latest'.format(
                        algo_history_loc))
                sys.exit(1)
            if 's3://' in algo_history_loc:
                history_s3_bucket = algo_history_loc.split('/')[-2]
                history_s3_key = algo_history_loc.split('/')[-1]
            elif 'redis://' in algo_history_loc:
                history_redis_key = algo_history_loc.split('/')[-1]
            elif 'file:/' in algo_history_loc:
                history_file = algo_history_loc.split(':')[-1]
            history_publish = True
        # end of parsing supported transport - trading history

        if args.algo_report_loc:
            algo_report_loc = args.algo_report_loc
            if ('file:/' not in algo_report_loc and
                    's3://' not in algo_report_loc and
                    'redis://' not in algo_report_loc):
                log.error(
                    'invalid -o <backtest dataset file> specified. '
                    '{} '
                    'please use either: '
                    '-o file:/opt/sa/tests/datasets/algo/SPY-latest.json or '
                    '-o s3://algoready/SPY-latest.json or '
                    '-o redis://SPY-latest'.format(
                        algo_report_loc))
                sys.exit(1)
            if 's3://' in algo_report_loc:
                report_s3_bucket = algo_report_loc.split('/')[-2]
                report_s3_key = algo_report_loc.split('/')[-1]
            elif 'redis://' in algo_report_loc:
                report_redis_key = algo_report_loc.split('/')[-1]
            elif 'file:/' in algo_report_loc:
                report_file = algo_report_loc.split(':')[-1]
            report_publish = True
        # end of parsing supported transport - trading performance report

        if args.algo_extract_loc:
            algo_extract_loc = args.algo_extract_loc
            if ('file:/' not in algo_extract_loc and
                    's3://' not in algo_extract_loc and
                    'redis://' not in algo_extract_loc):
                log.error(
                    'invalid -e <backtest dataset file> specified. '
                    '{} '
                    'please use either: '
                    '-e file:/opt/sa/tests/datasets/algo/SPY-latest.json or '
                    '-e s3://algoready/SPY-latest.json or '
                    '-e redis://SPY-latest'.format(
                        algo_extract_loc))
                sys.exit(1)
            if 's3://' in algo_extract_loc:
                extract_s3_bucket = algo_extract_loc.split('/')[-2]
                extract_s3_key = algo_extract_loc.split('/')[-1]
            elif 'redis://' in algo_extract_loc:
                extract_redis_key = algo_extract_loc.split('/')[-1]
            elif 'file:/' in algo_extract_loc:
                extract_file = algo_extract_loc.split(':')[-1]
            extract_publish = True
        # end of parsing supported transport - extract algorithm-ready

        use_custom_algo = True
    # end of set up for backtest

    if use_custom_algo:

        if args.run_on_engine:
            run_on_engine = True
            log.info('starting algo on the engine')
        else:
            log.info('starting algo')

        algo_res = run_custom_algo.run_custom_algo(
            mod_path=args.run_algo_in_file,
            ticker=ticker,
            balance=use_balance,
            commission=use_commission,
            start_date=use_start_date,
            end_date=use_end_date,
            config_file=use_config_file,
            name=use_name,
            auto_fill=auto_fill,
            config_dict=config_dict,
            load_from_s3_bucket=load_from_s3_bucket,
            load_from_s3_key=load_from_s3_key,
            load_from_redis_key=load_from_redis_key,
            load_from_file=load_from_file,
            load_compress=load_compress,
            load_publish=load_publish,
            load_config=load_config,
            report_redis_key=report_redis_key,
            report_s3_bucket=report_s3_bucket,
            report_s3_key=report_s3_key,
            report_file=report_file,
            report_compress=report_compress,
            report_publish=report_publish,
            report_config=report_config,
            history_redis_key=history_redis_key,
            history_s3_bucket=history_s3_bucket,
            history_s3_key=history_s3_key,
            history_file=history_file,
            history_compress=history_compress,
            history_publish=history_publish,
            history_config=history_config,
            extract_redis_key=extract_redis_key,
            extract_s3_bucket=extract_s3_bucket,
            extract_s3_key=extract_s3_key,
            extract_file=extract_file,
            extract_save_dir=extract_save_dir,
            extract_compress=extract_compress,
            extract_publish=extract_publish,
            extract_config=extract_config,
            publish_to_slack=publish_to_slack,
            publish_to_s3=publish_to_s3,
            publish_to_redis=publish_to_redis,
            dataset_type=dataset_type,
            serialize_datasets=serialize_datasets,
            compress=compress,
            encoding=encoding,
            redis_enabled=redis_enabled,
            redis_key=redis_key,
            redis_address=redis_address,
            redis_db=redis_db,
            redis_password=redis_password,
            redis_expire=redis_expire,
            redis_serializer=redis_serializer,
            redis_encoding=redis_encoding,
            s3_enabled=s3_enabled,
            s3_key=s3_key,
            s3_address=s3_address,
            s3_bucket=s3_bucket_name,
            s3_access_key=s3_access_key,
            s3_secret_key=s3_secret_key,
            s3_region_name=s3_region_name,
            s3_secure=s3_secure,
            slack_enabled=slack_enabled,
            slack_code_block=slack_code_block,
            slack_full_width=slack_full_width,
            dataset_publish_extract=extract_publish,
            dataset_publish_history=history_publish,
            dataset_publish_report=report_publish,
            run_on_engine=run_on_engine,
            auth_url=broker_url,
            backend_url=backend_url,
            include_tasks=include_tasks,
            ssl_options=ssl_options,
            transport_options=transport_options,
            path_to_config_module=path_to_config_module,
            timeseries=use_timeseries,
            trade_strategy=use_trade_strategy,
            verbose=verbose)

        show_label = 'algo.name={}'.format(
            use_name)
        show_extract = '{}'.format(
            algo_extract_loc)
        show_history = '{}'.format(
            algo_history_loc)
        show_report = '{}'.format(
            algo_report_loc)
        base_label = (
            'load={} '
            'extract={} '
            'history={} '
            'report={}'.format(
                args.run_algo_in_file,
                show_extract,
                show_history,
                show_report))
        show_label = (
            '{} running in engine task_id={} {}'.format(
                ticker,
                algo_res['rec'].get(
                    'task_id',
                    'missing-task-id'),
                base_label))
        if not run_on_engine:
            algo_trade_history_recs = algo_res['rec'].get(
                'history',
                [])
            show_label = (
                '{} algo.name={} {} trade_history_len={}'.format(
                    ticker,
                    use_name,
                    base_label,
                    len(algo_trade_history_recs)))
        if args.debug:
            log.info(
                'algo_res={}'.format(
                    algo_res))
            if algo_res['status'] == ae_consts.SUCCESS:
                log.info(
                    '{} - done running {}'.format(
                        ae_consts.get_status(status=algo_res['status']),
                        show_label))
            else:
                log.error(
                    '{} - done running {}'.format(
                        ae_consts.get_status(status=algo_res['status']),
                        show_label))
        else:
            if algo_res['status'] == ae_consts.SUCCESS:
                log.info(
                    '{} - done running {}'.format(
                        ae_consts.get_status(status=algo_res['status']),
                        show_label))
            else:
                log.error(
                    'run_custom_algo returned error: {}'.format(
                        algo_res['err']))
                sys.exit(1)
        # end of running the custom algo handler

        if mode == ae_consts.SA_MODE_EXTRACT:
            log.info('done extracting dataset - {}'.format(ticker))
        elif mode == ae_consts.SA_MODE_RUN_ALGO:
            log.info('done running algo - {}'.format(ticker))

        sys.exit(0)
    # end of handling mode-specific arg assignments

    # sanity checking the work and task are valid
    if not valid:
        log.error(
            'usage error: missing a supported mode: '
            '-f (for prepare a dataset) ')
        sys.exit(1)
    if required_task and not task_name:
        log.error(
            'usage error: missing a supported task_name')
        sys.exit(1)
    # end of sanity checks

    work['ticker'] = ticker
    work['ticker_id'] = ticker_id
    work['s3_bucket'] = s3_bucket_name
    work['s3_key'] = s3_key
    work['redis_key'] = redis_key
    work['s3_access_key'] = s3_access_key
    work['s3_secret_key'] = s3_secret_key
    work['s3_region_name'] = s3_region_name
    work['s3_address'] = s3_address
    work['s3_secure'] = s3_secure
    work['redis_address'] = redis_address
    work['redis_password'] = redis_password
    work['redis_db'] = redis_db
    work['redis_expire'] = redis_expire
    work['s3_enabled'] = s3_enabled
    work['redis_enabled'] = redis_enabled
    work['debug'] = debug
    work['label'] = 'ticker={}'.format(
        ticker)

    task_res = None
    if ae_consts.is_celery_disabled():
        work['celery_disabled'] = True
        log.debug(
            'starting without celery work={}'.format(
                ae_consts.ppj(work)))
        if mode == ae_consts.SA_MODE_PREPARE:
            task_res = prep_dataset.prepare_pricing_dataset(
                work)

        if debug:
            log.info(
                'done - result={} '
                'task={} status={} '
                'err={} label={}'.format(
                    ae_consts.ppj(task_res),
                    task_name,
                    ae_consts.get_status(status=task_res['status']),
                    task_res['err'],
                    work['label']))
        else:
            log.info(
                'done - result '
                'task={} status={} '
                'err={} label={}'.format(
                    task_name,
                    ae_consts.get_status(status=task_res['status']),
                    task_res['err'],
                    work['label']))

        if task_res['status'] == ae_consts.SUCCESS:
            image_res = None
            label = work['label']
            ticker = work['ticker']
            if plot_action == ae_consts.PLOT_ACTION_SHOW:
                log.info(
                    'showing plot')
                """
                minute_key = '{}_minute'.format(
                    redis_key)
                minute_df_res = build_df.build_df_from_redis(
                    label=label',
                    address=redis_address,
                    db=redis_db,
                    key=minute_key)

                minute_df = None
                if (
                        minute_df_res['status'] == SUCCESS
                        and minute_df_res['rec']['valid_df']):
                    minute_df = minute_df_res['rec']['data']
                    print(minute_df.columns.values)
                    column_list = [
                        'close',
                        'date'
                    ]
                """
                today_str = datetime.datetime.now().strftime(
                    '%Y-%m-%d')
                extract_req = work
                extract_req['redis_key'] = '{}_minute'.format(
                    work['redis_key'])
                extract_status, minute_df = \
                    extract_utils.extract_minute_dataset(
                        work_dict=work)
                if extract_status == ae_consts.SUCCESS:
                    log.info(
                        '{} - ticker={} creating chart date={}'.format(
                            label,
                            ticker,
                            today_str))
                    """
                    Plot Pricing with the Volume Overlay:
                    """
                    image_res = ae_charts.plot_overlay_pricing_and_volume(
                        log_label=label,
                        ticker=ticker,
                        date_format=ae_consts.IEX_MINUTE_DATE_FORMAT,
                        df=minute_df,
                        show_plot=True)

                    """
                    Plot the High-Low-Open-Close Pricing:
                    """
                    """
                    image_res = ae_charts.plot_hloc_pricing(
                        log_label=label,
                        ticker=ticker,
                        title='{} - Minute Pricing - {}'.format(
                            ticker,
                            today_str),
                        df=minute_df,
                        show_plot=True)
                    """

                    """
                    Plot by custom columns in the DataFrame
                    """
                    """
                    column_list = minute_df.columns.values
                    column_list = [
                        'date',
                        'close',
                        'high',
                        'low',
                        'open'
                    ]
                    image_res = ae_charts.plot_df(
                        log_label=label,
                        title='Pricing Title',
                        column_list=column_list,
                        df=minute_df,
                        xcol='date',
                        xlabel='Date',
                        ylabel='Pricing',
                        show_plot=True)
                    """
            elif plot_action == ae_consts.PLOT_ACTION_SAVE_TO_S3:
                log.info(
                    'coming soon - support to save to s3')
            elif plot_action == ae_consts.PLOT_ACTION_SAVE_AS_FILE:
                log.info(
                    'coming soon - support to save as file')
            if image_res:
                log.info(
                    '{} show plot - status={} err={}'.format(
                        label,
                        ae_consts.get_status(image_res['status']),
                        image_res['err']))
    else:
        log.info(
            'connecting to broker={} backend={}'.format(
                broker_url,
                backend_url))

        # Get the Celery app
        app = get_celery_app.get_celery_app(
            name=__name__,
            auth_url=broker_url,
            backend_url=backend_url,
            path_to_config_module=path_to_config_module,
            ssl_options=ssl_options,
            transport_options=transport_options,
            include_tasks=include_tasks)

        log.info(
            'calling task={} - work={}'.format(
                task_name,
                ae_consts.ppj(work)))
        job_id = app.send_task(
            task_name,
            (work,))
        log.info(
            'calling task={} - success job_id={}'.format(
                task_name,
                job_id))
    # end of if/else

# end of run_aws_backup


if __name__ == '__main__':
    run_aws_backup()



"""
Get defaults fields for supported teyps of data
"""

import analysis_engine.iex.consts as iex_consts
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def get_default_fields(
            field):
    """get_default_fields

    :param field: types of data to get
    """
    use_field = str(field).lower()
    if use_field == 'daily' or field == iex_consts.DATAFEED_DAILY:
        return ['KEY', 'date']
    elif use_field == 'quote' or field == iex_consts.DATAFEED_QUOTE:
        return ['KEY', 'date', 'minute']
    elif use_field == 'stats' or field == iex_consts.DATAFEED_STATS:
        return ['KEY']
    elif use_field == 'peers' or field == iex_consts.DATAFEED_PEERS:
        return ['KEY', 'peer']
    elif use_field == 'news' or field == iex_consts.DATAFEED_NEWS:
        return ['KEY', 'datetime']
    elif use_field == 'financials' or field == iex_consts.DATAFEED_FINANCIALS:
        return ['KEY', 'reportDate']
    elif use_field == 'earnings' or field == iex_consts.DATAFEED_EARNINGS:
        return ['KEY', 'EPSReportDate']
    elif use_field == 'dividends' or field == iex_consts.DATAFEED_DIVIDENDS:
        return ['KEY', 'exDate']
    elif use_field == 'company' or field == iex_consts.DATAFEED_COMPANY:
        return ['KEY']
    else:
        log.error(
            'get_default_fields({}) is not a supported '
            'field'.format(
                field))
        raise NotImplemented
# end of get_default_fields



"""
Dataset Publishing API
"""

import json
import boto3
import redis
import zlib
import analysis_engine.consts as ae_consts
import analysis_engine.compress_data as compress_data
import analysis_engine.set_data_in_redis_key as redis_utils
import analysis_engine.send_to_slack as slack_utils
import analysis_engine.write_to_file as file_utils
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def publish(
        data,
        label=None,
        convert_to_json=False,
        is_df=False,
        output_file=None,
        df_compress=False,
        compress=False,
        redis_enabled=True,
        redis_key=None,
        redis_address=None,
        redis_db=None,
        redis_password=None,
        redis_expire=None,
        redis_serializer='json',
        redis_encoding='utf-8',
        s3_enabled=True,
        s3_key=None,
        s3_address=None,
        s3_bucket=None,
        s3_access_key=None,
        s3_secret_key=None,
        s3_region_name=None,
        s3_secure=False,
        slack_enabled=False,
        slack_code_block=False,
        slack_full_width=False,
        verbose=False,
        silent=False,
        **kwargs):
    """publish

    Publish ``data`` to multiple optional endpoints:
    - a local file path (``output_file``)
    - minio (``s3_bucket`` and ``s3_key``)
    - redis (``redis_key``)
    - slack

    :return: status value
    :param data: data to publish
    :param convert_to_json: convert ``data`` to a
        json-serialized string. this function will throw if
        ``json.dumps(data)`` fails
    :param is_df: convert ``pd.DataFrame`` using
        ``pd.DataFrame.to_json()`` to a
        json-serialized string. this function will throw if
        ``to_json()`` fails
    :param label: log tracking label
    :param output_file: path to save the data
        to a file
    :param df_compress: optional - compress data that is a
        ``pandas.DataFrame`` before publishing
    :param compress: optional - compress before publishing
        (default is ``False``)
    :param verbose: optional - boolean to log output
        (default is ``False``)
    :param silent: optional - boolean no log output
        (default is ``False``)
    :param kwargs: optional - future argument support

    **(Optional) Redis connectivity arguments**

    :param redis_enabled: bool - toggle for auto-caching all
        datasets in Redis
        (default is ``True``)
    :param redis_key: string - key to save the data in redis
        (default is ``None``)
    :param redis_address: Redis connection string format: ``host:port``
        (default is ``localhost:6379``)
    :param redis_db: Redis db to use
        (default is ``0``)
    :param redis_password: optional - Redis password
        (default is ``None``)
    :param redis_expire: optional - Redis expire value
        (default is ``None``)
    :param redis_serializer: not used yet - support for future
        pickle objects in redis
    :param redis_encoding: format of the encoded key in redis

    **(Optional) Minio (S3) connectivity arguments**

    :param s3_enabled: bool - toggle for auto-archiving on Minio (S3)
        (default is ``True``)
    :param s3_key: string - key to save the data in redis
        (default is ``None``)
    :param s3_address: Minio S3 connection string format: ``host:port``
        (default is ``localhost:9000``)
    :param s3_bucket: S3 Bucket for storing the artifacts
        (default is ``dev``) which should be viewable on a browser:
        http://localhost:9000/minio/dev/
    :param s3_access_key: S3 Access key
        (default is ``trexaccesskey``)
    :param s3_secret_key: S3 Secret key
        (default is ``trex123321``)
    :param s3_region_name: S3 region name
        (default is ``us-east-1``)
    :param s3_secure: Transmit using tls encryption
        (default is ``False``)

    **(Optional) Slack arguments**

    :param slack_enabled: optional - boolean for
        publishing to slack
    :param slack_code_block: optional - boolean for
        publishing as a code black in slack
    :param slack_full_width: optional - boolean for
        publishing as a to slack using the full
        width allowed
    """

    status = ae_consts.NOT_RUN
    use_data = data
    if (
            not df_compress and
            not is_df and
            not use_data):
        log.info('missing data')
        return ae_consts.INVALID

    if convert_to_json and not is_df:
        if verbose:
            log.debug('start convert to json')
        use_data = json.dumps(data)
        if verbose:
            log.debug('done convert to json')
    if is_df:
        if verbose:
            log.debug('start df to_json')
        use_data = data.to_json(
            orient='records',
            date_format='iso')
        if verbose:
            log.debug('done df to_json')

    already_compressed = False
    if df_compress:
        use_data = compress_data.compress_data(
            data=data)
        already_compressed = True
    elif compress and not df_compress:
        if verbose:
            log.debug('compress start')
        use_data = zlib.compress(
            use_data.encode(
                redis_encoding), 9)
        already_compressed = True
        if verbose:
            log.debug('compress end')

    num_bytes = len(use_data)
    num_mb = ae_consts.get_mb(num_bytes)

    if verbose:
        log.debug(
            'start - file={} s3_key={} redis_key={} slack={} '
            'compress={} size={}MB'.format(
                output_file,
                s3_key,
                redis_key,
                slack_enabled,
                compress,
                num_mb))

    if s3_enabled and s3_address and s3_bucket and s3_key:
        endpoint_url = 'http://{}'.format(
            s3_address)
        if s3_secure:
            endpoint_url = 'https://{}'.format(
                s3_address)

        if verbose:
            log.debug(
                's3 start - {} endpoint_url={} '
                'region={}'.format(
                    label,
                    endpoint_url,
                    s3_region_name))

        s3 = boto3.resource(
            's3',
            endpoint_url=endpoint_url,
            aws_access_key_id=s3_access_key,
            aws_secret_access_key=s3_secret_key,
            region_name=s3_region_name,
            config=boto3.session.Config(
                signature_version='s3v4')
        )

        if s3.Bucket(s3_bucket) not in s3.buckets.all():
            if verbose:
                log.debug(
                    's3 creating bucket={} {}'.format(
                        s3_bucket,
                        label))
            s3.create_bucket(
                Bucket=s3_bucket)

        if verbose:
            log.debug(
                's3 upload start - bytes={} to {}:{} {}'.format(
                    num_mb,
                    s3_bucket,
                    s3_key,
                    label))

        s3.Bucket(
            s3_bucket).put_object(
                Key=s3_key,
                Body=use_data)

        if verbose:
            log.debug(
                's3 upload done - bytes={} to {}:{} {}'.format(
                    num_mb,
                    s3_bucket,
                    s3_key,
                    label))

    # end of s3_enabled

    if redis_enabled and redis_address and redis_key:
        redis_split = redis_address.split(':')
        redis_host = redis_split[0]
        redis_port = int(redis_split[1])
        log.debug(
            '{} redis={}:{}@{} connect '
            'key={} expire={}'.format(
                label if label else '',
                redis_host,
                redis_port,
                redis_db,
                redis_key,
                redis_expire))

        rc = redis.Redis(
            host=redis_host,
            port=redis_port,
            password=redis_password,
            db=redis_db)

        redis_res = redis_utils.set_data_in_redis_key(
            label=label,
            client=rc,
            key=redis_key,
            data=use_data,
            already_compressed=already_compressed,
            serializer=redis_serializer,
            encoding=redis_encoding,
            expire=redis_expire,
            px=None,
            nx=False,
            xx=False)

        if redis_res['status'] != ae_consts.SUCCESS:
            if verbose:
                log.debug('redis failed - {} {}'.format(
                    ae_consts.get_status(status=redis_res['status']),
                    redis_res['err']))
            return ae_consts.REDIS_FAILED
    # end of redis_enabled

    if output_file:
        if verbose:
            log.debug('file start - output_file={}'.format(
                output_file))
        file_exists = file_utils.write_to_file(
            output_file=output_file,
            data=data)
        if not file_exists:
            if verbose:
                log.debug('file failed - did not find output_file={}'.format(
                    output_file))
            return ae_consts.FILE_FAILED
        if verbose:
            log.debug('file done - output_file={}'.format(
                output_file))
    # end of writing to file

    if slack_enabled:
        if verbose:
            log.debug('slack start')
        slack_utils.post_success(
            msg=use_data,
            block=slack_code_block,
            full_width=slack_full_width)
        if verbose:
            log.debug('slack end')
    # end of sending to slack

    status = ae_consts.SUCCESS

    if verbose:
        log.debug(
            'end - {} file={} s3_key={} redis_key={} slack={} '
            'compress={} size={}MB'.format(
                ae_consts.get_status(status=status),
                output_file,
                s3_key,
                redis_key,
                slack_enabled,
                compress,
                num_mb))

    return status
# end of publish



"""
Fetch data from
`IEX <https://iexcloud.io/>`__
with the factory method ``fetch_data``
"""

import analysis_engine.iex.consts as iex_consts
import analysis_engine.iex.fetch_api as fetch_api
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def fetch_data(
        work_dict,
        fetch_type=None):
    """fetch_data

    Factory method for fetching data from
    IEX using an enum or string alias. Returns
    a pandas ``DataFrame`` and only supports
    one ticker at a time.

    Supported enums from: ``analysis_engine.iex.consts``

    ::

        fetch_type = iex_consts.FETCH_DAILY
        fetch_type = iex_consts.FETCH_MINUTE
        fetch_type = iex_consts.FETCH_QUOTE
        fetch_type = iex_consts.FETCH_STATS
        fetch_type = iex_consts.FETCH_PEERS
        fetch_type = iex_consts.FETCH_NEWS
        fetch_type = iex_consts.FETCH_FINANCIALS
        fetch_type = iex_consts.FETCH_EARNINGS
        fetch_type = iex_consts.FETCH_DIVIDENDS
        fetch_type = iex_consts.FETCH_COMPANY

    Supported ``work_dict['ft_type']`` string values:

    ::

        work_dict['ft_type'] = 'daily'
        work_dict['ft_type'] = 'minute'
        work_dict['ft_type'] = 'quote'
        work_dict['ft_type'] = 'stats'
        work_dict['ft_type'] = 'peers'
        work_dict['ft_type'] = 'news'
        work_dict['ft_type'] = 'financials'
        work_dict['ft_type'] = 'earnings'
        work_dict['ft_type'] = 'dividends'
        work_dict['ft_type'] = 'company'

    :param work_dict: dictionary of args for the pEX call
    :param fetch_type: optional - name or enum of the fetcher to create
                       can also be a lower case string
                       in work_dict['ft_type']
    """
    use_fetch_name = None
    if not fetch_type:
        fetch_type = work_dict.get(
            'ft_type',
            None)
    if fetch_type:
        use_fetch_name = str(fetch_type).lower()

    log.debug(
        f'name={use_fetch_name} '
        f'type={fetch_type} args={work_dict}')

    if (
            use_fetch_name == 'daily' or
            fetch_type == iex_consts.FETCH_DAILY):
        return fetch_api.fetch_daily(
            work_dict=work_dict)
    elif (
            use_fetch_name == 'minute' or
            fetch_type == iex_consts.FETCH_MINUTE):
        return fetch_api.fetch_minute(
            work_dict=work_dict)
    elif (
            use_fetch_name == 'quote' or
            fetch_type == iex_consts.FETCH_QUOTE):
        return fetch_api.fetch_quote(
            work_dict=work_dict)
    elif (
            use_fetch_name == 'stats' or
            fetch_type == iex_consts.FETCH_STATS):
        return fetch_api.fetch_stats(
            work_dict=work_dict)
    elif (
            use_fetch_name == 'peers' or
            fetch_type == iex_consts.FETCH_PEERS):
        return fetch_api.fetch_peers(
            work_dict=work_dict)
    elif (
            use_fetch_name == 'news' or
            fetch_type == iex_consts.FETCH_NEWS):
        return fetch_api.fetch_news(
            work_dict=work_dict)
    elif (
            use_fetch_name == 'financials' or
            fetch_type == iex_consts.FETCH_FINANCIALS):
        return fetch_api.fetch_financials(
            work_dict=work_dict)
    elif (
            use_fetch_name == 'earnings' or
            fetch_type == iex_consts.FETCH_EARNINGS):
        return fetch_api.fetch_earnings(
            work_dict=work_dict)
    elif (
            use_fetch_name == 'dividends' or
            fetch_type == iex_consts.FETCH_DIVIDENDS):
        return fetch_api.fetch_dividends(
            work_dict=work_dict)
    elif (
            use_fetch_name == 'company' or
            fetch_type == iex_consts.FETCH_COMPANY):
        return fetch_api.fetch_company(
            work_dict=work_dict)
    else:
        log.error(
            f'label={work_dict.get("label", None)} - '
            f'unsupported fetch_data('
            f'work_dict={work_dict}, '
            f'fetch_type={fetch_type}'
            f')')
        raise NotImplemented
    # end of supported fetchers
# end of fetch_data



"""
Parse data from TD

Supported environment variables:

::

    export DEBUG_TD_DATA=1

"""

import datetime
import copy
import analysis_engine.consts as ae_consts
import analysis_engine.build_result as build_result
import analysis_engine.api_requests as api_requests
import analysis_engine.td.fetch_data as td_fetch_data
import analysis_engine.td.consts as td_consts
import analysis_engine.work_tasks.publish_pricing_update as publisher
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def get_data_from_td(
        work_dict):
    """get_data_from_td

    Get pricing data from Tradier

    :param work_dict: request dictionary
    """
    label = 'get_data_from_td'

    log.debug(
        'task - {} - start '
        'work_dict={}'.format(
            label,
            work_dict))

    rec = {
        'data': None,
        'updated': None
    }
    res = {
        'status': ae_consts.NOT_RUN,
        'err': None,
        'rec': rec
    }

    ticker = None
    field = None
    ft_type = None

    try:

        ticker = work_dict.get(
            'ticker',
            ae_consts.TICKER)
        field = work_dict.get(
            'field',
            'daily')
        ft_type = work_dict.get(
            'ft_type',
            None)
        ft_str = str(ft_type).lower()
        label = work_dict.get(
            'label',
            label)
        orient = work_dict.get(
            'orient',
            'records')

        td_req = None
        if ft_type == td_consts.FETCH_TD_CALLS or ft_str == 'tdcalls':
            ft_type == td_consts.FETCH_TD_CALLS
            td_req = api_requests.build_td_fetch_calls_request(
                label=label)
        elif ft_type == td_consts.FETCH_TD_PUTS or ft_str == 'tdputs':
            ft_type == td_consts.FETCH_TD_PUTS
            td_req = api_requests.build_td_fetch_puts_request(
                label=label)
        else:
            log.error(
                '{} - unsupported ft_type={} ft_str={} ticker={}'.format(
                    label,
                    ft_type,
                    ft_str,
                    ticker))
            raise NotImplemented
        # if supported fetch request type

        clone_keys = [
            'ticker',
            's3_address',
            's3_bucket',
            's3_key',
            'redis_address',
            'redis_db',
            'redis_password',
            'redis_key'
        ]

        for k in clone_keys:
            td_req[k] = work_dict.get(
                k,
                '{}-missing-in-{}'.format(
                    k,
                    label))
        # end of cloning keys

        if not td_req:
            err = (
                '{} - ticker={} did not build an TD request '
                'for work={}'.format(
                    label,
                    td_req['ticker'],
                    work_dict))
            log.error(err)
            res = build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=rec)
            return res
        else:
            log.debug(
                '{} - ticker={} field={} '
                'orient={} fetch'.format(
                    label,
                    td_req['ticker'],
                    field,
                    orient))
        # if invalid td request

        df = None
        try:
            if 'from' in work_dict:
                td_req['from'] = datetime.datetime.strptime(
                    '%Y-%m-%d %H:%M:%S',
                    work_dict['from'])
            status_df, df = td_fetch_data.fetch_data(
                work_dict=td_req,
                fetch_type=ft_type)

            if status_df == ae_consts.SUCCESS:
                rec['data'] = df.to_json(
                    orient=orient)
                rec['updated'] = datetime.datetime.utcnow().strftime(
                    '%Y-%m-%d %H:%M:%S')
            elif status_df == ae_consts.EMPTY:
                res = build_result.build_result(
                    status=ae_consts.ERR,
                    err=(
                        f'did not fetch any data'),
                    rec=rec)
                return res
            else:
                err = (
                    '{} - ticker={} td_fetch_data.fetch_data field={} '
                    'failed fetch_data'
                    ''.format(
                        label,
                        td_req['ticker'],
                        ft_type))
                log.critical(err)
                res = build_result.build_result(
                    status=ae_consts.ERR,
                    err=err,
                    rec=rec)
                return res
        except Exception as f:
            err = (
                '{} - ticker={} field={} failed fetch_data '
                'with ex={}'.format(
                    label,
                    td_req['ticker'],
                    ft_type,
                    f))
            log.critical(err)
            res = build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=rec)
            return res
        # end of try/ex

        if ae_consts.ev('DEBUG_TD_DATA', '0') == '1':
            log.debug(
                '{} ticker={} field={} data={} to_json'.format(
                    label,
                    td_req['ticker'],
                    field,
                    rec['data']))
        else:
            log.debug(
                '{} ticker={} field={} to_json'.format(
                    label,
                    td_req['ticker'],
                    field))
        # end of if/else found data

        upload_and_cache_req = copy.deepcopy(td_req)
        upload_and_cache_req['celery_disabled'] = True
        upload_and_cache_req['data'] = rec['data']
        if not upload_and_cache_req['data']:
            upload_and_cache_req['data'] = '{}'
        use_field = field
        if use_field == 'news':
            use_field = 'news1'
        if 'redis_key' in work_dict:
            upload_and_cache_req['redis_key'] = '{}_{}'.format(
                work_dict.get(
                    'redis_key',
                    td_req['redis_key']),
                use_field)
        if 's3_key' in work_dict:
            upload_and_cache_req['s3_key'] = '{}_{}'.format(
                work_dict.get(
                    's3_key',
                    td_req['s3_key']),
                use_field)

        try:
            update_res = publisher.run_publish_pricing_update(
                work_dict=upload_and_cache_req)
            update_status = update_res.get(
                'status',
                ae_consts.NOT_SET)
            log.debug(
                '{} publish update status={} data={}'.format(
                    label,
                    ae_consts.get_status(status=update_status),
                    update_res))
        except Exception as f:
            err = (
                '{} - failed to upload td data={} to '
                'to s3_key={} and redis_key={}'.format(
                    label,
                    upload_and_cache_req,
                    upload_and_cache_req['s3_key'],
                    upload_and_cache_req['redis_key']))
            log.error(err)
        # end of try/ex to upload and cache

        if not rec['data']:
            log.debug(
                '{} - ticker={} no Tradier data field={} to publish'.format(
                    label,
                    td_req['ticker'],
                    field))
        # end of if/else

        res = build_result.build_result(
            status=ae_consts.SUCCESS,
            err=None,
            rec=rec)

    except Exception as e:
        res = build_result.build_result(
            status=ae_consts.ERR,
            err=(
                'failed - get_data_from_td '
                'dict={} with ex={}').format(
                    work_dict,
                    e),
            rec=rec)
    # end of try/ex

    log.debug(
        'task - get_data_from_td done - '
        '{} - status={} err={}'.format(
            label,
            ae_consts.get_status(res['status']),
            res['err']))

    return res
# end of get_data_from_td



"""
Date utils
"""

import datetime
import analysis_engine.consts as ae_consts


def last_close():
    """last_close

    Get last trading close time as a python ``datetime``

    How it works:

    - During market hours the returned ``datetime`` will be
        ``datetime.datetime.utcnow() - datetime.timedelta(hours=5)``
    - Before or after market hours, the returned ``datetime``
        will be 4:00 PM EST on the previous trading day which
        could be a Friday if this is called on a Saturday or Sunday.

    .. note:: does not detect holidays and non-trading
        days yet and assumes the system time is
        set to EST or UTC
    """
    now = (
        datetime.datetime.utcnow() - datetime.timedelta(hours=5))
    today = now.date()
    close = datetime.datetime(
        year=today.year,
        month=today.month,
        day=today.day,
        hour=16)
    market_start_time = datetime.datetime(
        year=today.year,
        month=today.month,
        day=today.day,
        hour=9,
        minute=30,
        second=0)
    market_end_time = datetime.datetime(
        year=today.year,
        month=today.month,
        day=today.day,
        hour=16,
        minute=0,
        second=0)

    if today.weekday() == 5:
        return close - datetime.timedelta(days=1)
    elif today.weekday() == 6:
        return close - datetime.timedelta(days=2)
    elif market_start_time <= now <= market_end_time:
        return now
    else:
        if now.hour < 16:
            close -= datetime.timedelta(days=1)
            if close.weekday() == 5:  # saturday
                return close - datetime.timedelta(days=1)
            elif close.weekday() == 6:  # sunday
                return close - datetime.timedelta(days=2)
            return close
        return close
    # if/else
# end of last_close


def get_last_close_str(
        fmt=ae_consts.COMMON_DATE_FORMAT):
    """get_last_close_str

    Get the Last Trading Close Date as a string
    with default formatting ae_consts.COMMON_DATE_FORMAT
    (YYYY-MM-DD)

    :param fmt: optional output format (default
        ae_consts.COMMON_DATE_FORMAT)
    """
    return last_close().strftime(fmt)
# end of get_last_close_str


def utc_now_str(
        fmt=ae_consts.COMMON_TICK_DATE_FORMAT):
    """utc_now_str

    Get the UTC now as a string
    with default formatting ae_consts.COMMON_TICK_DATE_FORMAT
    (YYYY-MM-DD HH:MM:SS)

    :param fmt: optional output format (default
        ae_consts.COMMON_TICK_DATE_FORMAT)
    """
    return datetime.datetime.utcnow().strftime(
        fmt)
# end of utc_now_str


def utc_date_str(
        fmt=ae_consts.COMMON_DATE_FORMAT):
    """utc_date_str

    Get the UTC date as a string
    with default formatting ``COMMON_DATE_FORMAT``

    :param fmt: optional output format (default
        COMMON_DATE_FORMAT ``YYYY-MM-DD``)
    """
    return datetime.datetime.utcnow().strftime(
        fmt)
# end of utc_date_str


def get_date_from_str(
        date_str,
        fmt=ae_consts.COMMON_TICK_DATE_FORMAT):
    """get_date_from_str

    Convert a date to a string where the
    default date formatting is ``ae_consts.COMMON_TICK_DATE_FORMAT``

    :param date_str: string date value with a format of ``fmt``
    :param fmt: date format ``YYYY-MM-DD HH:MM:SS`` by default
    """

    return datetime.datetime.strptime(
        date_str,
        fmt)
# end of get_date_from_str


def get_trade_open_xticks_from_date_col(
        date_list):
    """get_trade_open_xticks_from_date_col

    Call this to plot date strings in order
    with just the trading open as the xticks

    :param date_list: column from the ``pandas.DataFrame``
        like ``date_list=df['minute']``
    """

    open_of_trading_fmt = '%Y-%m-%d 09:30:00'

    date_strings = []
    date_labels = []

    last_date = None
    final_date = None
    for idx, date in enumerate(date_list):
        new_day = False
        final_date = date
        if not last_date:
            last_date = date
            next_open_of_trading = datetime.datetime.strptime(
                (date + datetime.timedelta(days=1)).strftime(
                    open_of_trading_fmt),
                ae_consts.COMMON_TICK_DATE_FORMAT)
            new_day = True
        else:
            if date > last_date and date > next_open_of_trading:
                new_day = True
                last_date = date
                next_open_of_trading = datetime.datetime.strptime(
                    (date + datetime.timedelta(days=1)).strftime(
                        open_of_trading_fmt),
                    ae_consts.COMMON_TICK_DATE_FORMAT)
        if new_day:
            date_strings.append(
                date.strftime(ae_consts.COMMON_TICK_DATE_FORMAT))
            date_labels.append(
                date.strftime('%m-%d %H:30'))
        # end of adding new day
    # end of for all dates to find the opens

    if final_date:
        should_add_last_point = False
        if final_date > last_date:
            should_add_last_point = True

        if should_add_last_point:
            date_strings.append(
                final_date.strftime(ae_consts.COMMON_TICK_DATE_FORMAT))
            date_labels.append(
                final_date.strftime('%m-%d %H:%M'))
        # end of adding final point
    # end of adding final point

    return date_strings, date_labels
# end of get_trade_open_xticks_from_date_col


def convert_epoch_to_datetime_string(
        epoch,
        fmt=ae_consts.COMMON_TICK_DATE_FORMAT,
        use_utc=True):
    """convert_epoch_to_datetime_string

    :param epoch: integer epoch time value
    :param fmt: optional string date format
    :param use_utc: if utc or local time - default is ``True``
    """

    if use_utc:
        return datetime.datetime.utcfromtimestamp(epoch).strftime(fmt)
    else:
        return datetime.datetime.fromtimestamp(epoch).strftime(fmt)
# end of convert_epoch_to_datetime_string


def epoch_to_dt(
        epoch,
        use_utc=False,
        convert_to_est=True):
    """epoch_to_dt

    Convert epoch milliseconds to datetime

    :param epoch: integer milliseconds
    :param use_utc: boolean to convert from ``UTC``
        default is ``False``
    :param convert_to_est: boolean to convert from
        ``UTC`` to ``EST``
    """

    converted_time = None
    if use_utc:
        converted_time = datetime.datetime.utcfromtimestamp(
            epoch)
    else:
        converted_time = datetime.datetime.fromtimestamp(
            epoch)
    # if/else

    if convert_to_est:
        converted_time = (
            converted_time - datetime.timedelta(hours=5))

    return converted_time
# end of epoch_to_dt


def get_days_between_dates(
        from_historical_date,
        last_close_to_use=None):
    """get_days_between_dates

    :param from_historical_date: historical date in time to start walking
                                 forward until the last close datetime
    :param last_close_to_use: starting date in time (left leg of window)
    """
    use_last_close = last_close_to_use
    if not use_last_close:
        use_last_close = last_close()

    dates = []
    while from_historical_date < last_close_to_use:
        dates.append(from_historical_date)
        from_historical_date += datetime.timedelta(
            days=1)
    return dates
# end of get_days_between_dates



import os
import sys
import warnings
import unittest

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

try:
    from distutils.command.build_py import build_py_2to3 as build_py
except ImportError:
    from distutils.command.build_py import build_py

"""
https://packaging.python.org/guides/making-a-pypi-friendly-readme/
check the README.rst works on pypi as the
long_description with:
twine check dist/*
"""
long_description = open('README.rst').read()

cur_path, cur_script = os.path.split(sys.argv[0])
os.chdir(os.path.abspath(cur_path))

requires_that_fail_on_rtd = [
    'awscli',
    'h5py',
    'keras',
    'scikit-learn',
    'tables',
    'ta-lib',
    'tensorflow'
]

install_requires = []

cur_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(
        cur_dir, 'requirements.txt'), encoding='utf-8') as f:
    install_requires = f.read().split()

# if not on readthedocs.io or travis ci get all the pips:
if (os.getenv('READTHEDOCS', '') == ''
        and os.getenv('TRAVIS', '') == ''):
    install_requires = install_requires + requires_that_fail_on_rtd

if sys.version_info < (3, 5):
    warnings.warn(
        'Less than Python 3.5 is not supported.',
        DeprecationWarning)


def analysis_engine_test_suite():
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover('tests', pattern='test_*.py')
    return test_suite


# Don't import analysis_engine module here, since deps may not be installed
sys.path.insert(
    0,
    os.path.join(
        os.path.dirname(__file__),
        'analysis_engine'))

setup(
    name='stock-analysis-engine',
    cmdclass={'build_py': build_py},
    version='1.7.8',
    description=(
        'Backtest 1000s of minute-by-minute '
        'trading algorithms. Automated '
        'pricing data ingestion from: '
        'IEX Cloud (https://iexcloud.io), '
        'Tradier (https://tradier.com) and '
        'FinViz. Datasets and trading '
        'performance automatically compressed and published '
        'to S3 for building AI training datasets '
        'for teaching DNNs how to trade. '
        'Runs on Kubernetes and docker-compose. '
        '>150 million trading history rows generated from +5000 algorithms'
        ''),
    long_description=long_description,
    author='Jay Johnson',
    author_email='jay.p.h.johnson@gmail.com',
    url='https://github.com/AlgoTraders/stock-analysis-engine',
    packages=[
        'analysis_engine',
        'analysis_engine.ai',
        'analysis_engine.iex',
        'analysis_engine.finviz',
        'analysis_engine.mocks',
        'analysis_engine.indicators',
        'analysis_engine.log',
        'analysis_engine.scripts',
        'analysis_engine.td',
        'analysis_engine.work_tasks',
        'analysis_engine.yahoo'
    ],
    package_data={},
    install_requires=install_requires,
    test_suite='setup.analysis_engine_test_suite',
    tests_require=[
    ],
    scripts=[
        'analysis_engine/scripts/aws_backup.py',
        'analysis_engine/scripts/fetch_new_stock_datasets.py',
        'analysis_engine/scripts/plot_history_from_local_file.py',
        'analysis_engine/scripts/publish_from_s3_to_redis.py',
        'analysis_engine/scripts/publish_ticker_aggregate_from_s3.py',
        'analysis_engine/scripts/run_backtest_and_plot_history.py',
        'analysis_engine/scripts/sa.py',
        'analysis_engine/scripts/start_algo.py',
        'analysis_engine/scripts/train_dnn_from_history.py',
        'tools/logs-dataset-collection.sh',
        'tools/logs-jupyter.sh',
        'tools/logs-workers.sh',
        'tools/run-algo-history-to-file.sh',
        'tools/run-algo-history-to-s3.sh',
        'tools/run-algo-report-to-file.sh',
        'tools/run-algo-report-to-s3.sh',
        'tools/ssh-jupyter.sh',
        'tools/ssh-workers.sh',
        'tools/update-stack.sh'
    ],
    entry_points={
        'console_scripts': [
            'sa = sa:run_sa_tool',
            (
                'fetch = fetch_new_stock_datasets'
                ':fetch_new_stock_datasets'),
            (
                'plot-history = '
                'plot_history_from_local_file'
                ':plot_local_history_file'),
            (
                'bt = run_backtest_and_plot_history'
                ':start_backtest_with_plot_history'),
            (
                'ae = start_algo'
                ':start_algo'),
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ])
